@@ -15,7 +15,7 @@
  *
  */
 /*
- *  Copyright (C) 2000-2014 Thomas Habets <thomas@habets.se>
+ *  Copyright (C) 2000-2015 Thomas Habets <thomas@habets.se>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -65,6 +65,14 @@
 #include <sys/types.h>
 #endif
 
+#if HAVE_SYS_PARAM_H
+#include <sys/param.h>
+#endif
+
+#if HAVE_GRP_H
+#include <grp.h>
+#endif
+
 #if HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
 #endif
@@ -129,7 +137,7 @@ arping_lookupdev_default(uint32_t srcip, uint32_t dstip, char *ebuf);
 
 static const char *version = VERSION; /* from autoconf */
 
-static libnet_t *libnet = 0;
+libnet_t *libnet = 0;
 
 /* Timestamp of last packet sent.
  * Used for timing, and assumes that reply is due to most recent sent query.
@@ -137,13 +145,13 @@ static libnet_t *libnet = 0;
 static struct timespec lastpacketsent;
 
 /* target string */
-static char *target = "huh? bug in arping?";
+static const char *target = "huh? bug in arping?";
 
 /*
  * Ping IP mode:   cmdline target
  * Ping MAC mode:  255.255.255.255, override with -T
  */
-static uint32_t dstip;
+uint32_t dstip;
 
 /*
  * Ping IP mode:   ethxmas, override with -t
@@ -151,9 +159,12 @@ static uint32_t dstip;
  */
 static uint8_t dstmac[ETH_ALEN];
 
-static uint32_t srcip;            /* autodetected, override with -S/-b/-0 */
+uint32_t srcip;                   /* autodetected, override with -S/-b/-0 */
 static uint8_t srcmac[ETH_ALEN];  /* autodetected, override with -s */
 
+static int16_t vlan_tag = -1; /* 802.1Q tag to add to packets. -V */
+static int16_t vlan_prio = -1; /* 802.1p prio to use with 802.1Q. -Q */
+
 static int beep = 0;                 /* beep when reply is received. -a */
 static int reverse_beep = 0;         /* beep when expected reply absent. -e */
 static int alsototal = 0;            /* print sent as well as received. -u */
@@ -165,8 +176,8 @@ static int finddup = 0;              /* finddup mode. -d */
 static int dupfound = 0;             /* set to 1 if dup found */
 static char lastreplymac[ETH_ALEN];  /* if last different from this then dup */
 
-static unsigned int numsent = 0;            /* packets sent */
-static unsigned int numrecvd = 0;           /* packets received */
+unsigned int numsent = 0;                   /* packets sent */
+unsigned int numrecvd = 0;                  /* packets received */
 static unsigned int max_replies = UINT_MAX; /* exit after -C replies */
 static unsigned int numdots = 0;            /* dots that should be printed */
 static const char* timestamp_type = NULL;   /* Incoming packet measurement ts type (-m) */
@@ -187,6 +198,7 @@ static enum { NORMAL,      /* normal output */
 
 static const uint8_t ethnull[ETH_ALEN] = {0, 0, 0, 0, 0, 0};
 static const uint8_t ethxmas[ETH_ALEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
+static const char* ip_broadcast = "255.255.255.255";
 
 int verbose = 0;  /* Increase with -v */
 
@@ -323,6 +335,10 @@ drop_privileges()
 /**
  * Do pcap_open_live(), except by using the pcap_create() interface
  * introduced in 2008 (libpcap 0.4) where available.
+ * This is so that we can set some options, which can't be set with
+ * pcap_open_live:
+ * 1) Immediate mode -- this prevents pcap from buffering.
+ * 2) Set timestamp type -- specify what type of timer to use.
  *
  * FIXME: Use pcap_set_buffer_size()?
  */
@@ -352,7 +368,7 @@ try_pcap_open_live(const char *device, int snaplen,
 
 #ifdef HAVE_PCAP_SET_IMMEDIATE_MODE
         // Without immediate mode some architectures (e.g. Linux with
-        // TPACKET_V3) will buffer replies and incorrectly upwards of
+        // TPACKET_V3) will buffer replies and incorrectly report upwards of
         // hundreds of milliseconds of delay.
         if ((rc = pcap_set_immediate_mode(pcap, 1))) {
                 if (verbose) {
@@ -456,11 +472,8 @@ do_pcap_open_live(const char *device, int snaplen,
  */
 void
 strip_newline(char* s) {
-        if (!*s) {
-                return;
-        }
         size_t n;
-        for (n = strlen(s); s[n - 1] == '\n'; --n) {
+        for (n = strlen(s); n && (s[n - 1] == '\n'); --n) {
                 s[n - 1] = 0;
         }
 }
@@ -495,9 +508,13 @@ do_libnet_init(const char *ifname, int recursive)
                         /* Sometimes libnet guesses an interface that it then
                          * can't use. Work around that by attempting to
                          * use "lo". */
-                        return do_libnet_init("lo", 1);
+                        do_libnet_init("lo", 1);
+                        if (libnet != NULL) {
+                                return;
+                        }
                 } else if (recursive) {
-                        /* Continue original execution. */
+                        /* Continue original execution to get that
+                         * error message. */
                         return;
                 }
                 fprintf(stderr, "arping: libnet_init(LIBNET_LINK, %s): %s\n",
@@ -548,7 +565,7 @@ getclock(struct timespec *ts)
  *
  */
 static char*
-format_mac(unsigned char* mac, char* buf, size_t bufsize) {
+format_mac(const unsigned char* mac, char* buf, size_t bufsize) {
         snprintf(buf, bufsize, "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x",
                 mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
         return buf;
@@ -595,6 +612,8 @@ extended_usage()
                "\n           Type of timestamp to use for incoming packets. Use -vv when\n"
                "           pinging to list available ones.\n"
 	       "    -q     Does not display messages, except error messages.\n"
+               "    -Q pri 802.1p priority to set. Should be used with 802.1Q (-V).\n"
+               "           Defaults to 0.\n"
 	       "    -r     Raw output: only the MAC/IP address is displayed for each reply.\n"
 	       "    -R     Raw output: Like -r but shows \"the other one\", can  be  combined\n"
 	       "           with -r.\n"
@@ -618,6 +637,7 @@ extended_usage()
 	       "           pinging MACs.\n"
 	       "    -U     Send unsolicited ARP.\n"
 	       "    -v     Verbose output. Use twice for more messages.\n"
+               "    -V num 802.1Q tag to add. Defaults to no VLAN tag.\n"
                "    -w     Time to wait between pings, in microseconds.\n"
                "    -W     Same as -w, but in floating point seconds.\n");
         printf("Report bugs to: thomas@habets.se\n"
@@ -640,7 +660,9 @@ standard_usage()
                "[ -T <host/ip ] "
                "[ -s <MAC> ] [ -t <MAC> ] [ -c <count> ]\n"
                "              "
-               "[ -C <count> ] [ -i <interface> ] [ -m <type> ] "
+               "[ -C <count> ] [ -i <interface> ] [ -m <type> ]\n"
+               "              "
+               "[ -V <vlan> ] [ -Q <priority> ] "
                "<host/ip/MAC | -B>\n");
 }
 
@@ -661,8 +683,11 @@ usage(int ret)
 }
 
 /**
+ * Check to see if it looks somewhat like a MAC address.
+ *
  * It was unclear from msdn.microsoft.com if their scanf() supported
  * [0-9a-fA-F], so I'll stay away from it.
+ *
  */
 static int is_mac_addr(const char *p)
 {
@@ -705,24 +730,31 @@ static int is_mac_addr(const char *p)
 }
 
 /**
- * lots of parms since C arrays suck
+ * parse mac address.
+ *
+ * return 1 on success.
  */
-static int get_mac_addr(const char *in,
-			unsigned int *n0,
-			unsigned int *n1,
-			unsigned int *n2,
-			unsigned int *n3,
-			unsigned int *n4,
-			unsigned int *n5)
+int
+get_mac_addr(const char *in, uint8_t *out)
 {
-	if (6 == sscanf(in, "%x:%x:%x:%x:%x:%x",n0,n1,n2,n3,n4,n5)) {
-		return 1;
-	} else if(6 == sscanf(in, "%2x%x.%2x%x.%2x%x",n0,n1,n2,n3,n4,n5)) {
-		return 1;
-	} else if(6 == sscanf(in, "%x-%x-%x-%x-%x-%x",n0,n1,n2,n3,n4,n5)) {
-		return 1;
-	}
-	return 0;
+        const char *formats[] = {
+                "%x:%x:%x:%x:%x:%x",
+                "%2x%x.%2x%x.%2x%x",
+                "%x-%x-%x-%x-%x-%x",
+                NULL,
+        };
+        int c;
+        for (c = 0; formats[c]; c++) {
+                unsigned int n[6];
+                if (6 == sscanf(in, formats[c],
+                                &n[0], &n[1], &n[2], &n[3], &n[4], &n[5])) {
+                        for (c = 0; c < 6; c++) {
+                                out[c] = n[c] & 0xff;
+                        }
+                        return 1;
+                }
+        }
+        return 0;
 }
 
 /**
@@ -836,17 +868,34 @@ pingmac_send(uint16_t id, uint16_t seq)
 			libnet_geterror(libnet));
 		sigint(0);
 	}
-	if (-1 == (eth = libnet_build_ethernet(dstmac,
-					       srcmac,
-					       ETHERTYPE_IP,
-					       NULL,
-					       0,
-					       libnet,
-					       eth))) {
-		fprintf(stderr, "libnet_build_ethernet(): %s\n",
-			libnet_geterror(libnet));
-		sigint(0);
-	}
+	if (vlan_tag >= 0) {
+                eth = libnet_build_802_1q(dstmac,
+                                          srcmac,
+                                          ETHERTYPE_VLAN,
+                                          vlan_prio,
+                                          0, // cfi
+                                          vlan_tag,
+                                          ETHERTYPE_IP,
+                                          NULL, // payload
+                                          0, // payload length
+                                          libnet,
+                                          eth);
+        } else {
+                eth = libnet_build_ethernet(dstmac,
+                                            srcmac,
+                                            ETHERTYPE_IP,
+                                            NULL, // payload
+                                            0, // payload length
+                                            libnet,
+                                            eth);
+        }
+        if (-1 == eth) {
+                fprintf(stderr, "arping: %s: %s\n",
+                        (vlan_tag >= 0) ? "libnet_build_802_1q()" :
+                        "libnet_build_ethernet()",
+                        libnet_geterror(libnet));
+                sigint(0);
+        }
 	if (verbose > 1) {
                 getclock(&lastpacketsent);
                 printf("arping: sending packet at time %ld.%09ld\n",
@@ -886,14 +935,32 @@ pingip_send()
 			libnet_geterror(libnet));
 		sigint(0);
 	}
-	if (-1 == (eth = libnet_build_ethernet(dstmac,
-					       srcmac,
-					       ETHERTYPE_ARP,
-					       NULL,
-					       0,
-					       libnet,
-					       eth))) {
-		fprintf(stderr, "arping: libnet_build_ethernet(): %s\n",
+
+        if (vlan_tag >= 0) {
+                eth = libnet_build_802_1q(dstmac,
+                                          srcmac,
+                                          ETHERTYPE_VLAN,
+                                          vlan_prio,
+                                          0, // cfi
+                                          vlan_tag,
+                                          ETHERTYPE_ARP,
+                                          NULL, // payload
+                                          0, // payload size
+                                          libnet,
+                                          eth);
+        } else {
+                eth = libnet_build_ethernet(dstmac,
+                                            srcmac,
+                                            ETHERTYPE_ARP,
+                                            NULL, // payload
+                                            0, // payload size
+                                            libnet,
+                                            eth);
+        }
+	if (-1 == eth) {
+		fprintf(stderr, "arping: %s: %s\n",
+			(vlan_tag >= 0) ? "libnet_build_802_1q()" :
+                        "libnet_build_ethernet()",
 			libnet_geterror(libnet));
 		sigint(0);
 	}
@@ -917,9 +984,11 @@ pingip_send()
  * \param h       packet metadata
  * \param packet  packet data
  */
-static void
+void
 pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
 {
+        const unsigned char *pkt_srcmac;
+        const struct libnet_802_1q_hdr *veth;
 	struct libnet_802_3_hdr *heth;
 	struct libnet_arp_hdr *harp;
         struct timespec arrival;
@@ -931,8 +1000,29 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
 
         getclock(&arrival);
 
-	heth = (void*)packet;
-	harp = (void*)((char*)heth + LIBNET_ETH_H);
+	if (vlan_tag >= 0) {
+		veth = (void*)packet;
+		harp = (void*)((char*)veth + LIBNET_802_1Q_H);
+		pkt_srcmac = veth->vlan_shost;
+	} else {
+                // Short packet.
+                if (h->caplen < LIBNET_ETH_H + LIBNET_ARP_H + 2*(ETH_ALEN + 4)) {
+                        return;
+                }
+
+		heth = (void*)packet;
+		harp = (void*)((char*)heth + LIBNET_ETH_H);
+		pkt_srcmac = heth->_802_3_shost;
+                // Wrong length of hardware address.
+                if (harp->ar_hln != ETH_ALEN) {
+                        return;
+                }
+
+                // Wrong length of protocol address.
+                if (harp->ar_pln != 4) {
+                        return;
+                }
+	}
 
         // ARP reply.
         if (htons(harp->ar_op) != ARPOP_REPLY) {
@@ -991,8 +1081,7 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
                 break;
         case NORMAL:
                 printf("%d bytes from %s (%s): index=%d",
-                       h->len, format_mac(heth->_802_3_shost,
-                                          buf, sizeof(buf)),
+                       h->len, format_mac(pkt_srcmac, buf, sizeof(buf)),
                        libnet_addr2name4(ip, 0), numrecvd);
 
                 if (alsototal) {
@@ -1004,16 +1093,14 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
         case QUIET:
                 break;
         case RAWRAW:
-                printf("%s %s", format_mac(heth->_802_3_shost,
-                                           buf, sizeof(buf)),
+                printf("%s %s", format_mac(pkt_srcmac, buf, sizeof(buf)),
                        libnet_addr2name4(ip, 0));
                 break;
         case RRAW:
                 printf("%s", libnet_addr2name4(ip, 0));
                 break;
         case RAW:
-                printf("%s", format_mac(heth->_802_3_shost,
-                                        buf, sizeof(buf)));
+                printf("%s", format_mac(pkt_srcmac, buf, sizeof(buf)));
                 break;
         default:
                 fprintf(stderr, "arping: can't happen!\n");
@@ -1029,11 +1116,11 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
         }
         if (numrecvd) {
                 if (memcmp(lastreplymac,
-                           heth->_802_3_shost, ETH_ALEN)) {
+                           pkt_srcmac, ETH_ALEN)) {
                         dupfound = 1;
                 }
         }
-        memcpy(lastreplymac, heth->_802_3_shost, ETH_ALEN);
+        memcpy(lastreplymac, pkt_srcmac, ETH_ALEN);
 
         numrecvd++;
         if (numrecvd >= max_replies) {
@@ -1046,9 +1133,12 @@ pingip_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
  * \param h       packet metadata
  * \param packet  packet data
  */
-static void
+void
 pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
 {
+        const unsigned char *pkt_dstmac;
+        const unsigned char *pkt_srcmac;
+        const struct libnet_802_1q_hdr *veth;
 	struct libnet_802_3_hdr *heth;
 	struct libnet_ipv4_hdr *hip;
 	struct libnet_icmpv4_hdr *hicmp;
@@ -1061,18 +1151,28 @@ pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
 
         getclock(&arrival);
 
-	heth = (void*)packet;
-	hip = (void*)((char*)heth + LIBNET_ETH_H);
-	hicmp = (void*)((char*)hip + LIBNET_IPV4_H);
+        if (vlan_tag >= 0) {
+                veth = (void*)packet;
+                hip = (void*)((char*)veth + LIBNET_802_1Q_H);
+                hicmp = (void*)((char*)hip + LIBNET_IPV4_H);
+                pkt_srcmac = veth->vlan_shost;
+                pkt_dstmac = veth->vlan_dhost;
+        } else {
+                heth = (void*)packet;
+                hip = (void*)((char*)heth + LIBNET_ETH_H);
+                hicmp = (void*)((char*)hip + LIBNET_IPV4_H);
+                pkt_srcmac = heth->_802_3_shost;
+                pkt_dstmac = heth->_802_3_dhost;
+        }
 
         // Dest MAC must be me.
-        if (memcmp(heth->_802_3_dhost, srcmac, ETH_ALEN)) {
+        if (memcmp(pkt_dstmac, srcmac, ETH_ALEN)) {
                 return;
         }
 
         // Source MAC must match, if set.
         if (memcmp(dstmac, ethxmas, ETH_ALEN)) {
-                if (memcmp(heth->_802_3_shost, dstmac, ETH_ALEN)) {
+                if (memcmp(pkt_srcmac, dstmac, ETH_ALEN)) {
                         return;
                 }
         }
@@ -1106,7 +1206,7 @@ pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
         case NORMAL:
                 printf("%d bytes from %s (%s): icmp_seq=%d time=%s", h->len,
                        libnet_addr2name4(*(int*)&hip->ip_src, 0),
-                       format_mac(heth->_802_3_shost, buf, sizeof(buf)),
+                       format_mac(pkt_srcmac, buf, sizeof(buf)),
                        htons(hicmp->icmp_seq),
                        ts2str(&lastpacketsent, &arrival, buf2, sizeof(buf2)));
                 break;
@@ -1114,11 +1214,11 @@ pingmac_recv(const char *unused, struct pcap_pkthdr *h, uint8_t *packet)
                 printf("%s", libnet_addr2name4(hip->ip_src.s_addr, 0));
                 break;
         case RRAW:
-                printf("%s", format_mac(heth->_802_3_shost, buf, sizeof(buf)));
+                printf("%s", format_mac(pkt_srcmac, buf, sizeof(buf)));
                 break;
         case RAWRAW:
                 printf("%s %s",
-                       format_mac(heth->_802_3_shost, buf, sizeof(buf)),
+                       format_mac(pkt_srcmac, buf, sizeof(buf)),
                        libnet_addr2name4(hip->ip_src.s_addr, 0));
                 break;
         default:
@@ -1240,6 +1340,11 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
                                        case DOT:
                                                printf(".");
                                                break;
+                                       case RAW:
+                                       case RAWRAW:
+                                       case RRAW:
+                                       case QUIET:
+                                               break;
                                        }
                                        fflush(stdout);
                                }
@@ -1281,19 +1386,40 @@ ping_recv(pcap_t *pcap, uint32_t packetwait, pcap_handler func)
        }
 }
 
+// return 1 on success.
+static int
+xresolve(libnet_t* l, const char *name, int r, uint32_t *addr)
+{
+        if (!strcmp(ip_broadcast, name)) {
+                *addr = 0xffffffff;
+                return 1;
+        }
+        *addr = libnet_name2addr4(l, (char*)name, r);
+        return *addr != 0xffffffff;
+}
+
 /**
  *
  */
-int main(int argc, char **argv)
+int
+arping_main(int argc, char **argv)
 {
 	char ebuf[LIBNET_ERRBUF_SIZE + PCAP_ERRBUF_SIZE];
 	char *cp;
 	int promisc = 0;
-	int srcip_given = 0;
-	int srcmac_given = 0;
+        const char *srcip_opt = NULL;
+        const char *dstip_opt = NULL;
+        // `dstip_given` can be set even when there's no arg past flags on the
+        // cmdline and -B not set. E.g. -d defaults to self, so requires no
+        // extra arg.
 	int dstip_given = 0;
+        const char *srcmac_opt = NULL;
+        const char *dstmac_opt = NULL;
 	const char *ifname = NULL;
-	char *parm;
+        int opt_B = 0;
+        int opt_T = 0;
+        int opt_U = 0;
+        const char *parm; // First argument, meaning the target IP.
 	int c;
 	unsigned int maxcount = -1;
 	int dont_use_arping_lookupdev=0;
@@ -1301,6 +1427,7 @@ int main(int argc, char **argv)
 	pcap_t *pcap;
 	enum { NONE, PINGMAC, PINGIP } mode = NONE;
 	unsigned int packetwait = 1000000;
+        char bpf_filter[64];
         ebuf[0] = 0;
 
         for (c = 1; c < argc; c++) {
@@ -1316,11 +1443,10 @@ int main(int argc, char **argv)
 	memcpy(dstmac, ethxmas, ETH_ALEN);
 
         while (EOF != (c = getopt(argc, argv,
-                                  "0aAbBC:c:dDeFhi:I:m:pPqrRs:S:t:T:uUvw:W:"))) {
+                                  "0aAbBC:c:dDeFhi:I:m:pPqQ:rRs:S:t:T:uUvV:w:W:"))) {
 		switch(c) {
 		case '0':
-			srcip = 0;
-			srcip_given = 1;
+			srcip_opt = "0.0.0.0";
 			break;
 		case 'a':
 			beep = 1;
@@ -1329,12 +1455,12 @@ int main(int argc, char **argv)
 			addr_must_be_same = 1;
 			break;
 		case 'b':
-			srcip = 0xffffffff;
-			srcip_given = 1;
+			srcip_opt = ip_broadcast;
 			break;
 		case 'B':
-			dstip = 0xffffffff;
-			dstip_given = 1;
+                        dstip_opt = ip_broadcast;
+                        dstip_given = 1;
+                        opt_B = 1;
 			break;
 		case 'c':
 			maxcount = atoi(optarg);
@@ -1381,6 +1507,15 @@ int main(int argc, char **argv)
 		case 'q':
 			display = QUIET;
 			break;
+                case 'Q':
+                        vlan_prio = atoi(optarg);
+                        if (vlan_prio < 0 || vlan_prio > 7) {
+                                fprintf(stderr,
+                                        "arping: 802.1p priority must be 0-7. It's %d\n",
+                                        vlan_prio);
+                                exit(1);
+                        }
+                        break;
 		case 'r':
 			display = (display==RRAW)?RAWRAW:RAW;
 			break;
@@ -1388,83 +1523,42 @@ int main(int argc, char **argv)
 			display = (display==RAW)?RAWRAW:RRAW;
 			break;
 		case 's': { /* spoof source MAC */
-			unsigned int n[6];
-			if (!get_mac_addr(optarg,
-					  &n[0],&n[1],&n[2],
-					  &n[3],&n[4],&n[5])){
-				fprintf(stderr, "arping: Weird MAC addr %s\n",
-					optarg);
-				exit(1);
-			}
-			for (c = 0; c < 6; c++) {
-				srcmac[c] = n[c] & 0xff;
-			}
-			srcmac_given = 1;
+                        srcmac_opt = optarg;
 			break;
 		}
 		case 'S': /* set source IP, may be null for don't-know */
-                        do_libnet_init(ifname, 0);
-			if (-1 == (srcip = libnet_name2addr4(libnet,
-							     optarg,
-							     LIBNET_RESOLVE))){
-				fprintf(stderr, "arping: Can't resolve %s, or "
-					"%s is broadcast. If it is, use -b"
-					" instead of -S\n", optarg,optarg);
-				exit(1);
-			}
-			srcip_given = 1;
+                        srcip_opt = optarg;
 			break;
 		case 't': { /* set taget mac */
-			unsigned int n[6];
-			if (mode == PINGMAC) {
-				fprintf(stderr, "arping: -t can only be used "
-					"in IP ping mode\n");
-				exit(1);
-			}
-			if (!get_mac_addr(optarg,
-					  &n[0],&n[1],&n[2],
-					  &n[3],&n[4],&n[5])){
-				fprintf(stderr, "Illegal MAC addr %s\n",
-					optarg);
-				exit(1);
-			}
-			for (c = 0; c < 6; c++) {
-				dstmac[c] = n[c] & 0xff;
-			}
+                        dstmac_opt = optarg;
 			mode = PINGIP;
 			break;
 		}
 		case 'T': /* set destination IP */
-			if (mode == PINGIP) {
-				fprintf(stderr, "arping: -T can only be used "
-					"in MAC ping mode\n");
-				exit(1);
-			}
-                        do_libnet_init(ifname, 0);
-			if (-1 == (dstip = libnet_name2addr4(libnet,
-							     optarg,
-							     LIBNET_RESOLVE))){
-				fprintf(stderr,"arping: Can't resolve %s, or "
-					"%s is broadcast. If it is, use -B "
-					"instead of -T\n",optarg,optarg);
-				exit(1);
-			}
+                        opt_T = 1;
+                        dstip_opt = optarg;
 			mode = PINGMAC;
 			break;
 		case 'u':
 			alsototal = 1;
 			break;
 		case 'U':
-			if (mode == PINGMAC) {
-				fprintf(stderr, "arping: -U can only be used "
-					"in IP ping mode\n");
-				exit(1);
-			}
+                        opt_U = 1;
 			unsolicited = 1;
+                        mode = PINGIP;
 			break;
 		case 'v':
 			verbose++;
 			break;
+		case 'V':
+			vlan_tag = atoi(optarg);
+                        if (vlan_tag < 0 || vlan_tag > 4095) {
+                                fprintf(stderr,
+                                        "arping: vlan tag must 0-4095. Is %d\n",
+                                        vlan_tag);
+                                exit(1);
+                        }
+                        break;
 		case 'w':
 			packetwait = (unsigned)atoi(optarg);
 			break;
@@ -1476,6 +1570,69 @@ int main(int argc, char **argv)
 		}
 	}
 
+        if (argc - optind > 1) {
+                // Can be zero if using -d or -B.
+                fprintf(stderr, "arping: Too many args on command line."
+                        " Expected at most one.\n");
+                exit(1);
+        }
+
+        if (((mode == PINGIP) && opt_T)
+            || (mode == PINGMAC) && (opt_B || dstmac_opt || opt_U)) {
+                fprintf(stderr, "arping: -T can only be used to ping MAC"
+                        " and -BtU only to ping IPs");
+                exit(1);
+        }
+        if (opt_T && opt_B) {
+                fprintf(stderr,
+                        "arping: -B can't be used with -T,"
+                        " since they set the same thing\n");
+                exit(1);
+        }
+
+        if (srcmac_opt != NULL) {
+                if (!get_mac_addr(srcmac_opt, srcmac)) {
+                        fprintf(stderr, "arping: Weird MAC addr %s\n",
+                                srcmac_opt);
+                        exit(1);
+                }
+        }
+
+        if (dstmac_opt != NULL) {
+                if (!get_mac_addr(dstmac_opt, dstmac)) {
+                        fprintf(stderr, "Illegal MAC addr %s\n", dstmac_opt);
+                        exit(1);
+                }
+        }
+
+        if (srcip_opt != NULL) {
+                do_libnet_init(ifname, 0);
+                if (!xresolve(libnet, srcip_opt, LIBNET_RESOLVE, &srcip)) {
+                        fprintf(stderr, "arping: Can't resolve %s, or "
+                                "%s is broadcast. If it is, use -b"
+                                " instead of -S\n", srcip_opt,srcip_opt);
+                        exit(1);
+                }
+        }
+
+        if (dstip_opt) {
+                do_libnet_init(ifname, 0);
+                if (!xresolve(libnet, dstip_opt, LIBNET_RESOLVE, &dstip)) {
+                        fprintf(stderr,"arping: Can't resolve %s, or "
+                                "%s is broadcast. If it is, use -B "
+                                "instead of -T\n",dstip_opt,dstip_opt);
+                        exit(1);
+                }
+        }
+
+        if (vlan_prio >= 0 && vlan_tag == -1) {
+                fprintf(stderr, "arping: -Q requires the use of 802.1Q (-V)\n");
+                exit(1);
+        }
+        if (vlan_prio == -1) {
+                vlan_prio = 0;
+        }
+
         if (verbose > 1) {
 #if HAVE_CLOCK_MONOTONIC
                 struct timespec ts;
@@ -1551,9 +1708,7 @@ int main(int argc, char **argv)
 				"\n");
 			exit(1);
 		}
-		if (-1 == (dstip = libnet_name2addr4(libnet,
-						     parm,
-						     LIBNET_RESOLVE))) {
+                if (!xresolve(libnet, parm, LIBNET_RESOLVE, &dstip)) {
 			fprintf(stderr, "arping: Can't resolve %s\n", parm);
 			exit(1);
 		}
@@ -1564,7 +1719,6 @@ int main(int argc, char **argv)
 	 * parse parm into dstmac
 	 */
 	if (mode == PINGMAC) {
-		unsigned int n[6];
 		if (optind + 1 != argc) {
 			usage(1);
 		}
@@ -1574,17 +1728,12 @@ int main(int argc, char **argv)
 				"argument\n");
 			exit(1);
 		}
-		if (!get_mac_addr(argv[optind],
-				  &n[0],&n[1],&n[2],
-				  &n[3],&n[4],&n[5])) {
+                if (!get_mac_addr(argv[optind], dstmac)) {
 			fprintf(stderr, "arping: Illegal mac addr %s\n",
 				argv[optind]);
 			return 1;
 		}
-		for (c = 0; c < 6; c++) {
-			dstmac[c] = n[c] & 0xff;
-		}
-	}	
+	}
 
 	target = parm;
 	/*
@@ -1671,14 +1820,28 @@ int main(int argc, char **argv)
 
 	if (mode == PINGIP) {
 		/* FIXME: better filter with addresses? */
-		if (-1 == pcap_compile(pcap, &bp, "arp", 0,-1)) {
-			fprintf(stderr, "arping: pcap_compile(): error\n");
+                if (vlan_tag >= 0) {
+                        snprintf(bpf_filter, sizeof(bpf_filter),
+                                 "vlan %u and arp", vlan_tag);
+                } else {
+                        snprintf(bpf_filter, sizeof(bpf_filter), "arp");
+                }
+                if (-1 == pcap_compile(pcap, &bp, bpf_filter, 0, -1)) {
+                        fprintf(stderr, "arping: pcap_compile(%s): error\n",
+                                bpf_filter);
 			exit(1);
 		}
 	} else { /* ping mac */
 		/* FIXME: better filter with addresses? */
-		if (-1 == pcap_compile(pcap, &bp, "icmp", 0,-1)) {
-			fprintf(stderr, "arping: pcap_compile(): error\n");
+                if (vlan_tag >= 0) {
+                        snprintf(bpf_filter, sizeof(bpf_filter),
+                                 "vlan %u and icmp", vlan_tag);
+                } else {
+                        snprintf(bpf_filter, sizeof(bpf_filter), "icmp");
+                }
+                if (-1 == pcap_compile(pcap, &bp, bpf_filter, 0,-1)) {
+                        fprintf(stderr, "arping: pcap_compile(%s): error\n",
+                                bpf_filter);
 			exit(1);
 		}
 	}
@@ -1690,7 +1853,7 @@ int main(int argc, char **argv)
 	/*
 	 * final init
 	 */
-	if (!srcmac_given) {
+        if (srcmac_opt == NULL) {
 		if (!(cp = (char*)libnet_get_hwaddr(libnet))) {
 			fprintf(stderr, "arping: libnet_get_hwaddr(): %s\n",
 				libnet_geterror(libnet));
@@ -1698,7 +1861,7 @@ int main(int argc, char **argv)
 		}
 		memcpy(srcmac, cp, ETH_ALEN);
 	}
-	if (!srcip_given) {
+        if (srcip_opt == NULL) {
 		if (-1 == (srcip = libnet_get_ipaddr4(libnet))) {
 			fprintf(stderr,
                                 "arping: Unable to get the IPv4 address of "
@@ -1785,4 +1948,6 @@ int main(int argc, char **argv)
  * c-basic-offset: 8
  * indent-tabs-mode: nil
  * End:
+ *
+ * vim: ts=8 sw=8
  */
